// ;-*-C++-*- *  Time-stamp: "2010-02-28 15:44:31 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *
 * License: GPL
 *
 * Initial version: 2008-09-02
 *
 * NeuroML import/export methods for CModel
 */

#include <string>
#include <iostream>
#include <boost/regex.hpp>

#include "config.h"

#include "model.h"


using namespace std;

#ifdef LIBXML_READER_ENABLED





int
CModel::import_NetworkML( const char *fname, bool appending)
{
	LIBXML_TEST_VERSION;

	xmlDoc *doc = xmlReadFile( fname, NULL, 0);
	if ( !doc )
		return CN_NMLIN_NOFILE;

	int retval = import_NetworkML( doc, fname, appending);

	xmlFreeDoc( doc);

	return retval;
}




static xmlNode* find_named_root_child_elem( xmlNode *node, const char *elem);

int
CModel::import_NetworkML( xmlDoc *doc, const char *fname, bool appending)
{
	int retval = 0;

	// we pass up on validation (for which we would need to keep a
	// .dtd or Schema at hand), and proceed to extracting elements

	xmlNode *root_node = xmlDocGetRootElement( doc),
		*n;

      // read meta:notes and make out a name for the model
	if ( !root_node ) {
		cerr << "Failed to obtain root element\n";
		retval = CN_NMLIN_NOELEM;
		goto out;
	}

      // give it a name: assume it's generated by neuroConstruct for now
	if ( !appending ) {
		reset();
		if ( !(n = find_named_root_child_elem( root_node, "notes")) )
			cerr << "<notes> element not found\n";
			// this is not critical, so just keep the user
			// informed and proceed
		else
			if ( n->type == XML_ELEMENT_NODE ) {  // only concern ourselves with nodes of this type
				xmlChar *notes_s = xmlNodeGetContent( n);
				// look for a substring specific to neuroConstruct, which is obviously speculative
				boost::regex pattern( ".*project: (\\w*).*");
				boost::cmatch found;
				name = boost::regex_match( (char*)notes_s, found, pattern)
					? name = found[1]
					: "(unmatched)";
				xmlFree( notes_s);
			} else
				name = "(unnamed)";
	}

	if ( verbosely > 1 )
		printf( "Model \"%s\": %sing topology from %s\n",
			name.c_str(), (appending ?"Append" :"Import"), fname);

	// In the following calls to _process_{populations,instances}
	// functions, the actual order of appearance of these nodes in
	// the xml file doesn't matter, thanks to the xml contents
	// being already wholly read and available to us as a tree.

      // process <populations>
	if ( !(n = find_named_root_child_elem( root_node, "populations")) ) {
		retval = CN_NMLIN_NOELEM;
		goto out;
	} // assume there is only one <populations> element: don't loop to catch more
	if ( (retval = _process_populations( n->children)) < 0)	// note n->children, which is in fact a pointer to the first child
		goto out;

      // process <projections>
      // don't strictly require any projections as long as there are some neurons
	if ( (n = find_named_root_child_elem( root_node, "projections")) ) {
		if ( (retval = _process_projections( n->children)) < 0 )
			goto out;
	} else
		if ( verbosely > 2 )
			cout << "No projections found\n";

out:
	// we are done with topology; now put units' variables on a vector
	finalize_additions();
	// can call time_step only after finalize_additions

	cout << endl;

	return retval;
}


static xmlNode*
find_named_root_child_elem( xmlNode *node,     // node to start search from
			    const char *elem)  // name of the element searched for
{
	xmlNode *n;

	for ( n = node->children; n; n = n->next ) {
		if ( n->type == XML_ELEMENT_NODE ) {
			if ( xmlStrEqual( n->name, BAD_CAST elem) )
				return n;
// the <populations> and <projections> nodes are expected to appear as
// direct children of the root node; so don't go search deeper than that

//			if ( n->children ) { // go search deeper
//				ni = find_named_elem( n->children, elem);
//				if ( ni )
//					return ni;
//			}
		}
	}
	return NULL;
}






int
CModel::_process_populations( xmlNode *n)
{
	xmlChar *group_id_s, *cell_type_s;

	int	pop_cnt = 0;

	for ( ; n; n = n->next ) // if is NULL (parent had no children), we won't do a single loop
		if ( n->type == XML_ELEMENT_NODE && xmlStrEqual( n->name, BAD_CAST "population") ) {

			group_id_s = xmlGetProp( n, BAD_CAST "name"); // BAD_CAST is just a cast to xmlChar*
								      // with a catch that libxml functions
								      // expect strings pointed to to be good UTF
			if ( !group_id_s ) {
				fprintf( stderr, "<population> element missing a \"name\" attribute near line %d\n", n->line);
				return CN_NMLIN_BADATTR;
			}
		      // probably having an unnamed popuation isn't an error so serious as to abort the
		      // operation, but discipline is above all

			cell_type_s = xmlGetProp( n, BAD_CAST "cell_type");
			// now we know the type of cells included in this population; remember it to pass on to
			// _process_population_instances, where it is used to select an appropriate unit type
			// when actually adding a neuron to the model

		      // but well, let's check if we have units of that species in stock
			if ( !unit_species_is_neuron((char*)cell_type_s) && !unit_family_is_neuron((char*)cell_type_s) ) {
				fprintf( stderr, "Bad cell species or family (\"%s\") in population \"%s\"\n",
					 (char*)cell_type_s, group_id_s);
				xmlFree( cell_type_s), xmlFree( group_id_s);
				return CN_NMLIN_BADCELLTYPE;
			}

			xmlNode *nin = n->children;  // again, ->children means ->first
			if ( nin )
				for ( ; nin; nin = nin->next )  // deal with multiple <instances> nodes
					if ( nin->type == XML_ELEMENT_NODE && xmlStrEqual( nin->name, BAD_CAST "instances") ) {
						int subretval = _process_population_instances( nin->children,
											       group_id_s, cell_type_s);
						if ( subretval < 0 ) {
							xmlFree( group_id_s);
							return subretval;
						}
						if ( verbosely > 2 )
							printf( " %5d instance(s) of type \"%s\" in population \"%s\"\n",
								subretval, cell_type_s,  group_id_s);
						pop_cnt++;
					}
			xmlFree( cell_type_s);
			xmlFree( group_id_s);
		}

	if ( verbosely > 1 )
		printf( "\tTotal %d population(s)\n", pop_cnt);

	return pop_cnt;
}






int
CModel::_process_projections( xmlNode *n)
{
	// much the same code as in _process_populations

	xmlChar *prj_name_s, *prj_src_s, *prj_tgt_s;

	size_t pop_cnt = 0;

	for ( ; n; n = n->next ) {
		if ( n->type != XML_ELEMENT_NODE || !xmlStrEqual( n->name, BAD_CAST "projection") )
			continue;

		prj_name_s = xmlGetProp( n, BAD_CAST "name");
		if ( !prj_name_s ) {
			cerr << "<projection> element missing a \"name\" attribute near line " << n->line << endl;
			return CN_NMLIN_BADATTR;
		}

		prj_src_s  = xmlGetProp( n, BAD_CAST "source");
		prj_tgt_s  = xmlGetProp( n, BAD_CAST "target");
		if ( !prj_src_s || !prj_tgt_s ) {
			cerr << "Projection \"" << prj_name_s << "\" missing a \"source\" and/or \"target\" attribute near line " << n->line << endl;
			xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
			return CN_NMLIN_BADATTR;
		}

		xmlNode *nin;
		nin = n->children;
		if ( !nin )
			cerr << "Empty <projection> node near line " << n->line << endl;

		xmlChar *synapse_type_s = NULL;

		for ( ; nin; nin = nin->next )
			if ( nin->type == XML_ELEMENT_NODE && xmlStrEqual( nin->name, BAD_CAST "synapse_props") ) {
				synapse_type_s = xmlGetProp( nin, BAD_CAST "synapse_type");
				if ( !unit_species_is_synapse((char*)synapse_type_s) &&
				     !unit_family_is_synapse((char*)synapse_type_s) ) {
					fprintf( stderr, "Bad synapse type \"%s\" near line %u\n",
						 (char*)synapse_type_s, nin->line);
					xmlFree( synapse_type_s), xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
					return CN_NMLIN_BADCELLTYPE;
				}
			}

		for ( nin = n->children; nin; nin = nin->next )
			if ( nin->type == XML_ELEMENT_NODE && xmlStrEqual( nin->name, BAD_CAST "connections") ) {
				int subretval = _process_projection_connections( nin->children,
										 prj_name_s, synapse_type_s,
										 prj_src_s, prj_tgt_s);
				if ( subretval < 0 ) {
					xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
					return subretval;
				}
				if ( verbosely > 2 )
					printf( " %5d connection(s) of type \"%s\" in projection \"%s\"\n",
						subretval, synapse_type_s,  prj_name_s);
				pop_cnt++;
			}
		xmlFree( prj_name_s), xmlFree( prj_src_s), xmlFree( prj_tgt_s);
	}

	if ( verbosely > 1 )
		printf( "\tTotal %zd projection(s)\n", pop_cnt);

	return (int)pop_cnt;
}







int
CModel::_process_population_instances( xmlNode *n, const xmlChar *group_prefix, const xmlChar *type_s)
{
	int	retval = 0;  // also keeps a count of added neurons

	double	x, y, z;
	char	cell_id[CN_MAX_LABEL_SIZE];

	xmlNode *nin;

	for ( ; n; n = n->next ) {
		if ( n->type != XML_ELEMENT_NODE || !xmlStrEqual( n->name, BAD_CAST "instance") )
			continue;

		xmlChar *id_s = xmlGetProp( n, BAD_CAST "id");
		if ( !id_s ) {
		      // could be less strict here and allow empty ids, which will then be composed
		      // from group_prefix + id (say, "LN0", "LN1" and so on); but then, as
		      // individual <projection>s would have to reference both endpoints by explicit
		      // ids, it is obviously prone to error to have <instance> ids depend solely on
		      // their order of appearance.
		      // So we bark at empty ids.
			cerr << "<instance> element without an \"id\" attribute near line " << n->line << endl;
			retval = CN_NMLIN_BADATTR;
			goto out;
		}

		size_t total_len = xmlStrlen( group_prefix) + xmlStrlen( id_s);
		if ( total_len >= CN_MAX_LABEL_SIZE ) {
			cerr << "<instance> combined label (" << group_prefix << id_s << ") exceeding " << CN_MAX_LABEL_SIZE << " near line " << n->line << endl;
			xmlFree( id_s);
			retval = CN_NMLIN_BIGLABEL;
			goto out;
		}
		_longest_label = max( _longest_label,
				      (short unsigned)snprintf( cell_id, CN_MAX_LABEL_SIZE-1, "%s.%s",
								group_prefix, id_s));  // here, a new instance is given a name

		xmlFree( id_s);

		if ( !(nin = n->children) )
			goto out;

		for ( ; nin; nin = nin->next ) {
			if ( !(nin->type == XML_ELEMENT_NODE &&
			       xmlStrEqual( nin->name, BAD_CAST "location")) )
				continue;

			xmlChar *x_s = xmlGetProp( nin, BAD_CAST "x"),
				*y_s = xmlGetProp( nin, BAD_CAST "y"),
			*z_s = xmlGetProp( nin, BAD_CAST "z");
		      // here we do actually insert neurons into the model
			if ( !(x_s && y_s && z_s) )
				fprintf( stderr, "<location> element missing full set of coordinates near line %d\n", nin->line);
				//retval = CN_NMLIN_BADATTR;

			x = strtod( (char*)x_s, NULL), y = strtod( (char*)y_s, NULL), z = strtod( (char*)z_s, NULL);
			int	add_retval = -1;
			C_BaseNeuron *neu = add_neuron_species( (char*)type_s, cell_id, false);

			if ( !neu || neu->_status & CN_UERROR ) {
				if ( neu )
					delete neu;
				add_retval = -1;
				fprintf( stderr, "Failed to add a neuron \"%s\" near line %u\n", cell_id, n->line);
				retval = CN_NMLIN_STRUCTERROR;
			} else {
				neu->_serial_id = _global_unit_id_reservoir++;
				retval++;
			}
			xmlFree( x_s), xmlFree( y_s), xmlFree( z_s);

			if ( retval < 0 )
				goto out;
		}
	}
out:
	return retval;
}




int
CModel::_process_projection_connections( xmlNode *n,
					 const xmlChar *synapse_name, const xmlChar *type_s,
					 const xmlChar *src_grp_prefix, const xmlChar *tgt_grp_prefix)
{
	// similar to _process_population_instances, except that we read some more attributes (source and
	// target units)

	int	retval = 0;  // is also a counter of synapses

	char	//synapse_id [CN_MAX_LABEL_SIZE],
		src_s[CN_MAX_LABEL_SIZE],
		tgt_s[CN_MAX_LABEL_SIZE];
	double	weight;

	C_BaseSynapse	*y;

	for ( ; n; n = n->next ) {
		if ( n->type != XML_ELEMENT_NODE || !xmlStrEqual( n->name, BAD_CAST "connection") )
			continue;

		xmlChar //*synapse_id_s  = xmlGetProp( n, BAD_CAST "id"),
			*src_cell_id_s = xmlGetProp( n, BAD_CAST "pre_cell_id"),
			*tgt_cell_id_s = xmlGetProp( n, BAD_CAST "post_cell_id"),
			*weight_s      = xmlGetProp( n, BAD_CAST "weight");
		if ( /*!synapse_id_s || */ !src_cell_id_s || !tgt_cell_id_s ) {
			fprintf( stderr, "A <connection> element without \"pre_cell_id\" and/or \"post_cell_id\" attribute near line %u\n", n->line);
			retval = CN_NMLIN_BADATTR;
			goto out;
		}

		snprintf( src_s, CN_MAX_LABEL_SIZE-1, "%s.%s", src_grp_prefix, src_cell_id_s);
		snprintf( tgt_s, CN_MAX_LABEL_SIZE-1, "%s.%s", tgt_grp_prefix, tgt_cell_id_s);

		if ( !weight_s ) {
			fprintf( stderr, "Assuming 0 for a synapse of \"%s.%s\" to \"%s%s\" without a \"weight\" attribute near line %u\n",
				 src_grp_prefix, src_cell_id_s, tgt_grp_prefix, tgt_cell_id_s, n->line);
			weight = 0.;
		}

		y = add_synapse_species( (char*)type_s, src_s, tgt_s, weight, true, false);

		if ( !y || y->_status & CN_UERROR ) {
			if ( y )
				delete y;
			fprintf( stderr, "Failed to add an \"%s\" synapse from \"%s\" to \"%s\" near line %u\n",
				 (char*)type_s, src_s, tgt_s, n->line);
			retval = CN_NMLIN_STRUCTERROR;
			goto out;
		} else
			retval++;


		/* xmlFree( synapse_id_s), */ xmlFree( src_cell_id_s), xmlFree( tgt_cell_id_s),
			xmlFree( weight_s);
		continue;
	out:
		/* xmlFree( synapse_id_s), */ xmlFree( src_cell_id_s), xmlFree( tgt_cell_id_s),
			xmlFree( weight_s);
		break;

	}

	return retval;
}



int
CModel::export_NetworkML( const char *fname)
{
	int retval = 0;

	LIBXML_TEST_VERSION;

	fprintf( stderr, "export_NetworkML() not implemented yet\n");

	return retval;
}







#else
#error Need an XMLREADER-enabled libxml2 (>2.6)

#endif // LIBXML_READER_ENABLED
// EOF
