// ;-*-C++-*- *  Time-stamp: "2010-02-19 16:17:43 hmmr"
/*
 * Author: Andrei Zavada (a.zavada@sussex.ac.uk)
 *
 * License: GPL
 *
 * Initial version: 2008-11-11
 *
 */



#include <vector>
#include <iostream>
#include <fstream>
#include <limits>
#include <cmath>
#include <cstdlib>
#include <cstring>

using namespace std;


typedef vector<double>::iterator vd_i;
typedef vector<unsigned>::iterator vu_i;


enum TConvType {
	SDFF_CMP_SQDIFF,
	SDFF_CMP_WEIGHT
};

enum TCFOpType {
	SDFF_CFOP_AVG,
	SDFF_CFOP_PROD,
	SDFF_CFOP_SUM
};


struct SOptions {
	const char
		*working_dir,
		*target_profiles_dir;
	vector<string>
		units;
	TCFOpType
		cf_op_type;
	vector<unsigned>
		dims;
	bool	go_sdf:1,
		use_shf:1,
		do_normalise:1,
		do_matrix_output:1,
		do_column_output:1,
		assume_no_shf_value:1;
	double	sample_from,
		sample_period,
		sample_window;
	unsigned
		field_n,
		of_fields,
		skipped_first_lines;
	TConvType
		conv_type;

	SOptions()
	      : working_dir ("."),
		target_profiles_dir ("."),
		cf_op_type (SDFF_CFOP_AVG),
		go_sdf (true),
		use_shf (false),
		do_normalise (false),
		do_matrix_output (true),
		do_column_output (false),
		assume_no_shf_value (false),
		sample_from (0),
		sample_period (0),
		sample_window (0),
		field_n (1),
		of_fields (1),
		skipped_first_lines (0),
		conv_type (SDFF_CMP_WEIGHT)
		{}
};

static SOptions Options;

static double *MxTgt = NULL,
	*M = NULL;

static size_t dim_prod = 1;


static int get_unit_cf( const char *unit_label, double *result);

static int parse_cmdline( int argc, char *argv[]);
static void usage( const char *argv0);

#define SDFCAT_EARGS		-1
#define SDFCAT_EHELPREQUEST	-2
#define SDFCAT_EFILES		-3
#define SDFCAT_ERANGES		-4


static int read_matrices_from_sxf( const char* fname, double* M, double* H, double *sdf_max_p = NULL);
static int construct_matrix_from_var( const char* fname, double* M);
static int read_matrix( const char*, double*);
static int write_matrix( const char*, double*);
static double convolute_matrix_against_target( double*);



int
main( int argc, char *argv[])
{
	int retval = 0;

	if ( argc == 1 ) {
		usage( argv[0]);
		return SDFCAT_EARGS;
	}

	{
		int parse_retval = parse_cmdline( argc, argv);
		if ( parse_retval ) {
			if ( parse_retval == SDFCAT_EHELPREQUEST )
				usage( argv[0]);
			return -1;
		}

		if ( Options.assume_no_shf_value && Options.use_shf ) {
			cerr << "Conflicting options (-H and -H-)\n";
			return -1;
		}
	}

      // cd as requested
	char *pwd = NULL;
	if ( Options.working_dir ) {
		pwd = getcwd( NULL, 0);
		if ( chdir( Options.working_dir) ) {
			fprintf( stderr, "Failed to cd to \"%s\"\n", Options.working_dir);
			return -2;
		}
	}


	double grand_CF = 0.;
	vector<double> unit_CFs;

	double *GrandTgt;

      // buffer for target profiles
	for ( vu_i I = Options.dims.begin(); I != Options.dims.end(); I++ )
		dim_prod *= *I;
	if ( !(MxTgt = new double[dim_prod]) ||
	     !(GrandTgt = new double[dim_prod]) ||
	     !(M = new double[dim_prod]) ) {
		cerr << "Failed to create target profile buffer\n";
		retval = -3;
		goto out;
	}

	for ( size_t i = 0; i < dim_prod; i++ )
		GrandTgt[i] = 0.;

      // collect individual CF values
	for ( vector<string>::iterator uI = Options.units.begin(); uI != Options.units.end(); uI++ ) {
		double ucf;
		if ( get_unit_cf( uI->c_str(), &ucf) ) {
			retval = -4;
			goto out;
		}
		unit_CFs.push_back( ucf);
		for ( size_t i = 0; i < dim_prod; i++ )
			GrandTgt[i] += M[i];
	}

      // calculate grand CF
	switch ( Options.cf_op_type ) {

	case SDFF_CFOP_SUM:
		grand_CF = 0.;
		for ( vd_i xI = unit_CFs.begin(); xI != unit_CFs.end(); xI++ )
			grand_CF += *xI;
	    break;

	case SDFF_CFOP_AVG:
		grand_CF = 0.;
		for ( vd_i xI = unit_CFs.begin(); xI != unit_CFs.end(); xI++ )
			grand_CF += *xI;
		grand_CF /= unit_CFs.size();
	    break;

	case SDFF_CFOP_PROD:
		grand_CF = 1.;
		for ( vd_i xI = unit_CFs.begin(); xI != unit_CFs.end(); xI++ )
			grand_CF *= *xI;
	    break;
	}

      // save CF
	{
		ofstream CFsaved( "CF");
		CFsaved << grand_CF << endl;
		CFsaved.close();
	}
      // and GrandTgt
	write_matrix( "overall.mx", GrandTgt);

	cout << grand_CF << endl;

      // shutdown
out:
	delete[] MxTgt;
	delete[] GrandTgt;
	delete[] M;

	if ( pwd )
		if ( chdir( pwd) )
			;

	return retval;
}














static int
get_unit_cf( const char* ulabel, double *result_p)
{
	int retval = 0;

	double *H = NULL;
	if ( (Options.use_shf && !(H = new double[dim_prod])) ) {
		cerr << "Failed on new double[sizeof(double) * " << dim_prod << "]\n";
		retval = -1;
		goto out;
	}

	if ( Options.go_sdf ) {
		if ( read_matrices_from_sxf( (string(ulabel) + ".sxf").c_str(), M, H) &&
		     (Options.assume_no_shf_value = true, Options.use_shf = false,
		      read_matrices_from_sxf( (string(ulabel) + ".sdf").c_str(), M, H)) ) {
			fprintf( stderr, "Failed to read data from`%s.s{x,d}f'\n", ulabel);
			retval = -2;
			goto out;
		}
	} else  // go var
		if ( construct_matrix_from_var( (string(ulabel) + ".var").c_str(), M) ) {
			fprintf( stderr, "Failed to read `%s.var'\n", ulabel);
			retval = -2;
			goto out;
		}

	if ( (Options.do_matrix_output || Options.do_column_output)
	     && Options.dims.size() == 2 ) {  // only applicable to 2-dim matrices
		write_matrix( (string(ulabel) + ".sdf").c_str(), M);
		if ( Options.use_shf )
			write_matrix( (string(ulabel) + ".shf").c_str(), H);
	}

	if ( read_matrix( (string(Options.target_profiles_dir) + '/' + ulabel + ".sdf.target").c_str(),
			  MxTgt) ) {
		fprintf( stderr, "Failed to read target profile for `%s'\n", ulabel);
		retval = -2;
		goto out;
	}

	if ( result_p )
		*result_p = convolute_matrix_against_target( M);

out:
	delete H;

	return retval;
}





// ------------------------- matrix io ------

static int
read_matrices_from_sxf( const char *fname, double *M, double *H, double *sdf_max_p)
{
	ifstream ins( fname);
	if ( !ins.good() ) {
//		cerr << "No results in " << fname << endl;
		return -1;
	}

	size_t	ignored_lines = 0;

	double	sdf_max = -INFINITY,
		_;
	size_t	idx, row;
	for ( idx = row = 0; idx < dim_prod; idx += (++row > Options.skipped_first_lines)) {
		while ( ins.peek() == '#' ) {
			ignored_lines++;
			ins.ignore( numeric_limits<streamsize>::max(), '\n');
		}

		if ( ins.eof() ) {
			fprintf( stderr, "Short read from %s at element %zd\n", fname, idx);
			return -2;
		}
		ins >> _;       // time
		ins >> M[idx];  // sdf
		if ( !Options.assume_no_shf_value )
			ins >> (Options.use_shf ? H[idx] : _);  // shf
		ins >> _;       // nspikes

		if ( M[idx] > sdf_max )
			sdf_max = M[idx];
	}

	if ( Options.use_shf )
		for ( idx = 0; idx < dim_prod; idx++ )
			M[idx] *= H[idx];

	if ( Options.do_normalise ) {
		for ( idx = 0; idx < dim_prod; idx++ )
			M[idx] /= sdf_max;
		// if ( H )
		// 	for ( idx = 0; idx < dim_prod; idx++ )
		// 		H[idx] /= sdf_max;
	}

	if ( sdf_max_p )
		*sdf_max_p = sdf_max;

	return 0;
}





static int
construct_matrix_from_var( const char *fname, double *M)
{
	ifstream ins( fname);
	if ( !ins.good() ) {
		cerr << "No results in " << fname << endl;
		return -1;
	}

	double	at, _, var;
	vector<double> sample;
	size_t	idx;

	for ( idx = 0; idx < dim_prod; idx++ ) {
		M[idx] = 0.;

		while ( ins.peek() == '#' )
			ins.ignore( numeric_limits<streamsize>::max(), '\n');

		if ( ins.eof() ) {
			fprintf( stderr, "Short read from %s at element %zd\n", fname, idx);
			return -2;
		}

		sample.clear();
		do {
			ins >> at;
			for ( unsigned f = 1; f <= Options.of_fields; f++ )
				if ( f == Options.field_n )
					ins >> var;
				else
					ins >> _;

			if ( at < Options.sample_from + Options.sample_period * idx - Options.sample_window/2 )
				continue;

			sample.push_back( var);

		} while ( !(at > Options.sample_from + Options.sample_period * idx + Options.sample_window/2) );

		for ( vd_i X = sample.begin(); X != sample.end(); X++ )
			M[idx] += *X;
		M[idx] /= sample.size();
//		cout << "at " << Options.sample_from + Options.sample_period * idx << " sample size "<< sample.size() << endl;
	}

	// if ( Options.do_normalise ) {
	// 	for ( idx = 0; idx < dim_prod; idx++ )
	// 		M[idx] /= sdf_max;
	// 	// if ( H )
	// 	// 	for ( idx = 0; idx < dim_prod; idx++ )
	// 	// 		H[idx] /= sdf_max;
	// }

	return 0;
}





static int
read_matrix( const char *fname, double *M)
{
	ifstream ins( fname);
	if ( !ins.good() ) {
		cerr << "No results in " << fname << endl;
		return -1;
	}

//	size_t	ignored_lines = 0;
	while ( ins.peek() == '#' ) {
		ins.ignore( numeric_limits<streamsize>::max(), '\n');  // skip header
//		ignored_lines++;
	}
//	cout << ignored_lines << " lines skipped\n";

	size_t	idx;
	for ( idx = 0; idx < dim_prod; idx++ )
		if ( ins.eof() ) {
			fprintf( stderr, "Short read from %s at element %zd\n", fname, idx);
			return -1;
		} else
			ins >> M[idx];
	return 0;
}






static int
write_matrix( const char *fname, double *X)
{
	if ( Options.do_matrix_output ) {
		ofstream outs( (string(fname) + ".mx").c_str());
		for ( size_t k = 0; k < Options.dims[0]; k++ )
			for ( size_t l = 0; l < Options.dims[1]; l++ ) {
				if ( l > 0 )  outs << "\t";
				outs << X[k*Options.dims[0] + l];
				if ( l == Options.dims[1]-1 ) outs << endl;
			}
		if ( !outs.good() )
			return -1;
	}

	if ( Options.do_column_output ) {
		ofstream outs( (string(fname) + ".col").c_str());
		for ( size_t k = 0; k < Options.dims[0]; k++ )
			for ( size_t l = 0; l < Options.dims[1]; l++ )
				outs << l << "\t" << k << "\t" << X[k*Options.dims[0] + l] << endl;
		if ( !outs.good() )
			return -1;
	}

	return 0;
}






static double
convolute_matrix_against_target( double *M)
{
	double	CF = 0.;
	size_t idx;

	switch ( Options.conv_type ) {
	case SDFF_CMP_WEIGHT:
		for ( idx = 0; idx < dim_prod; idx++ )
			CF += M[idx] * MxTgt[idx];
		break;
	case SDFF_CMP_SQDIFF:
		for ( idx = 0; idx < dim_prod; idx++ )
			CF += pow( M[idx] - MxTgt[idx], 2);
		CF = sqrt( CF);
		break;
	}

	return CF;
}









// ------------------------- service functions ----

static int
parse_cmdline( int argc, char *argv[])
{
	char c;
	while ( (c = getopt( argc, argv, "C:Rd:f:G:u:H::Nx:V:z:o:F:h")) != -1 ) {
		switch ( c ) {
		case 'C':	Options.working_dir = optarg;				break;

		case 'R':	Options.go_sdf = false;					break;

		case 'd':	if ( sscanf( optarg, "%lg:%lg:%lg",
					     &Options.sample_from, &Options.sample_period, &Options.sample_window) < 2 ) {
					cerr << "Expecting three parameter with -d (from:period[:window])\n";
					return SDFCAT_EARGS;
				}
				if ( Options.sample_window == 0. )
					Options.sample_window = Options.sample_period;	break;

		case 'f':	if ( sscanf( optarg, "%d:%d",
					     &Options.field_n, &Options.of_fields) < 1 ) {
					cerr << "Expecting two parameters with -f (field:fields)\n";
					return SDFCAT_EARGS;
				}							break;

		case 'G':	Options.target_profiles_dir = optarg;			break;

		case 'u':	Options.units.push_back( string(optarg));			break;

		case 'H':	if ( optarg )
					if ( strcmp( optarg, "-") == 0 )
						Options.assume_no_shf_value = true, Options.use_shf = false;
					else {
						cerr << "Unrecognised option to -H: `" << optarg << "\n";
						return SDFCAT_EARGS;
					}
				else
					Options.use_shf = true;				break;

		case 'N':	Options.do_normalise = true;				break;

		case 'V':	if ( strcmp( optarg, "sqdiff" ) == 0 )
					Options.conv_type = SDFF_CMP_SQDIFF;
				else if ( strcmp( optarg, "weight") == 0 )
					Options.conv_type = SDFF_CMP_WEIGHT;
				else {
					cerr << "-V takes `sqdiff' or `weight'\n";
					return SDFCAT_EARGS;
				}
			break;
		case 'z':	if ( strcmp( optarg, "sum" ) == 0 )
					Options.cf_op_type = SDFF_CFOP_SUM;
				else if ( strcmp( optarg, "avg") == 0 )
					Options.cf_op_type = SDFF_CFOP_AVG;
				else if ( strcmp( optarg, "prod") == 0 )
					Options.cf_op_type = SDFF_CFOP_PROD;
				else {
					cerr << "-X can be `sum', `avg' or `prod'\n";
					return SDFCAT_EARGS;
				}
			break;
		case 'o':	Options.do_matrix_output = (strchr( optarg, 'm') != NULL);
				Options.do_column_output = (strchr( optarg, 'c') != NULL);
			break;

		case 'x':
		{
			unsigned d;
			if ( sscanf( optarg, "%ud", &d) < 1 ) {
				cerr << "-x takes an unsigned\n";
				return SDFCAT_EARGS;
			}
			Options.dims.push_back( d);
		}   break;

		case 'F':
			if ( sscanf( optarg, "%ud", &Options.skipped_first_lines) < 1 ) {
				cerr << "-F takes an unsigned\n";
				return SDFCAT_EARGS;
			}
		    break;

		case 'h':
			return SDFCAT_EHELPREQUEST;
		default:
			return SDFCAT_EARGS;
		}
	}

	if ( Options.units.empty() ) {
		cerr << "No units (-u) specified\n";
		return SDFCAT_EARGS;
	}
	if ( Options.dims.empty() ) {
		cerr << "No dimensions (-x) specified\n";
		return SDFCAT_EARGS;
	}

	return 0;
}




static void
usage( const char *argv0)
{
	cout << "Usage: " << argv0 << "\n"
		" -C <dir>\t\tcd into dir before working\n"
		" -G <dir>\t\tSearch for target profiles in dir (default " << Options.target_profiles_dir << ")\n"
		" -u <label>\t\tUnit to process (e.g., PN.0; multiple entries as necessary;\n"
		"\t\t\t  will try label.sxf then label.sdf)\n"
		" -x <dim>\t\tDimensions for the target and data folded matrices\n"
		"\t\t\t  (repeat as necessary)\n"
		" -V[sqdiff|weight]\tConvolution method:\n"
		"\t\t\t  sum of squared differences between unit and target profiles,\n"
		"\t\t\t  or sum of unit profile values weighted by those in target profile (default)\n"
		" -z[sum|avg|prod]\tOperation applied to individual unit CF (default is avg)\n"
		"\n"
		" -R\t\t\tCollect .var data rather than .sxf\n"
		"With -R, use\n"
		" -f <unsigned n1>:<unsigned n2>\n"
		"\t\t\tExtract n1th field of n2 consec. fields per record\n"
		"\t\t\t  (default " << Options.field_n << " of " << Options.of_fields << ")\n"
		" -d <double f>:<double p>:<double ws>\tSample from time f at period p with window size ws\n"
		"otherwise:\n"
		" -F <unsigned>\t\tRead sxf data from that position, not from 0\n"
		" -H \t\t\tMultiply sdf by shf\n"
		" -H-\t\t\tAssume there is no shf field in .sxf file\n"
		"\n"
		" -o[mc]\t\t\tWrite <unit>.[m]atrix and/or .[c]ol profiles\n"
		" -h\t\t\tDisplay this help\n";
}

// EOF
